#[doc = include_str!("../README.md")]
use jsonschema::{Draft, JSONSchema};
use serde_json::{Map, Value};
use std::fmt::Display;

/// The JSON schema inspector (validator)
///
/// This contains the necessary glue to make all the schemas from the OpenAPI spec (tested on 3.0.3) work with the `jsonschema` crate.
/// This is achieved by simply resolving the local references present in the schemas with the actual schemas.
///
/// It is recommended to initialize one `SchemaInspector` for one schema only once, and use the same instance to validate/inspect
/// multiple JSON objects to avoid performance overhead of compiling the schema multiple times.
///
/// Note: Resolving references is optional and is not required for schemas that do not contain any references to nested schemas.
#[derive(Debug)]
pub struct SchemaInspector {
    schema: JSONSchema, // the compiled schema used to validate JSON
}

impl SchemaInspector {
    const PROP: &'static str = "properties";
    const REF: &'static str = "$ref";
    const AOF: &'static str = "allOf";

    /// Initialize the `SchemaInspector` using the name of the type provided to the function using the turbofish syntax instead of using a string
    ///
    /// Note: Resolving references is optional and is not required for schemas that do not contain any references to nested schemas.
    /// It is recommended to set `resolve_refs` to false when there are no references present in the schema to avoid extra overhead.
    ///
    /// ## Example
    /// ```
    /// use schema_police::{SchemaInspector, InitError};
    ///
    /// fn construct_inspector(schema: &serde_json::Value) -> Result<SchemaInspector, InitError> {
    ///     SchemaInspector::new_infer::<SchemaType>(schema, true)
    /// }
    ///
    /// #[derive(utoipa::ToSchema)]
    /// struct SchemaType {
    ///     field: String,
    /// }
    /// ```
    ///
    /// Note: the function internally uses [`std::any::type_name`] to get the name of the type. Accodring to the std lib docs, the function
    /// is meant to be used for debugging and makes no guarantees. Therefore, it is not recommended to rely on this method in mission critical contexts.
    pub fn new_infer<T>(schemas: &Value, resolve_refs: bool) -> Result<Self, InitError> {
        // get target name from type inference
        let target = std::any::type_name::<T>()
            .split("::")
            .last()
            .ok_or(InitError::InferenceFailure)?;
        Self::new(target, schemas, resolve_refs)
    }

    /// Construct an instance of the `SchemaInspector` using the provided OpenAPI spec's schemas (in the form of `serde_json::Value`)
    ///
    /// Note: Resolving references is optional and is not required for schemas that do not contain any references to nested schemas.
    /// It is recommended to set `resolve_refs` to false when there are no references present in the schema to avoid extra overhead.
    pub fn new(target: &str, schemas: &Value, resolve_refs: bool) -> Result<Self, InitError> {
        use InitError::*;
        // extract schemas object
        let Value::Object(schemas) = schemas else {
            return Err(InitError::InvalidSchema(None));
        };
        // extract target from schema
        let Value::Object(mut target) = schemas
            .get(target)
            .ok_or_else(|| SchemaNotFound(target.to_owned()))?
            .to_owned()
        else {
            return Err(InitError::InvalidSchema(None));
        };
        if resolve_refs {
            // extract properties
            let Value::Object(mut props) = target
                .remove(Self::PROP)
                .ok_or_else(|| InvalidSchema(Some(Self::PROP.to_owned())))?
            else {
                return Err(InvalidSchema(Some(Self::PROP.to_owned())));
            };
            Self::resolve_references(schemas, &mut props)?;
            // reinsert resolved properties
            target.insert(Self::PROP.into(), Value::Object(props));
        }
        Ok(Self {
            schema: JSONSchema::options()
                .with_draft(Draft::Draft202012)
                .compile(&Value::Object(target))
                .map_err(|_| CompileFailure)?,
        })
    }

    /// Replace internal '$ref' references with the correct schemas
    /// The current implementation disregards the paths in the references and only uses the schema names
    /// to retrieve them from the root schemas object
    fn resolve_references(
        schemas: &Map<String, Value>,
        props: &mut Map<String, Value>,
    ) -> Result<(), InitError> {
        let to_resolve: Vec<(String, String)> = props
            .iter()
            .filter_map(|(key, val)| {
                Self::extract_ref(val)
                    .or_else(|| {
                        // Get the first item of allOf array and get the "$ref" field from that object
                        val.get(Self::AOF)
                            .and_then(|aof| aof.get(0).and_then(Self::extract_ref))
                    })
                    .map(|schema_name| (key.to_owned(), schema_name.to_owned()))
            })
            .collect();
        // replace the references in props fields with the correct schemas
        for (field, schema) in &to_resolve {
            let inner = props.get_mut(field).expect("known field cannot fail");
            *inner = schemas
                .get(schema)
                .ok_or_else(|| InitError::ResolutionFailure(schema.to_owned()))?
                .to_owned();
            if let Some(Value::Object(inner_props)) = inner.get_mut(Self::PROP) {
                Self::resolve_references(schemas, inner_props)?;
            }
        }
        Ok(())
    }

    /// Extract the fragment and pack it with the given field name
    #[inline]
    fn extract_ref(val: &Value) -> Option<String> {
        let refr = val.get(Self::REF)?;
        let Value::String(final_refr) = refr else {
            return None;
        };
        final_refr.split('/').last().map(ToOwned::to_owned)
    }

    /// Inspect (validate) the given JSON value to check if it conforms to the compiled schema
    ///
    /// [`SchemaError`] is returned upon failure which contains a detailed explanation of everything
    /// that is inconsistent with the compiled schema.
    pub fn inspect(&self, json: &Value) -> Result<(), SchemaError> {
        self.schema.validate(json).map_err(|err| {
            let mut cause = err.fold(String::new(), |mut acc, err| {
                acc.push_str(&err.to_string());
                acc.push_str(", ");
                acc
            });
            cause.pop();
            cause.pop();
            SchemaError { cause }
        })
    }
}

/// The error that is returned by the [`SchemaInspector`] when the provided JSON to the `inspect` function
/// fails to conform to the schema the inspector was compiled with.
///
/// The error type internally holds a string containing detailed information about what part of the JSON differs from the schema
#[derive(Debug, Clone, PartialEq, Eq)]
pub struct SchemaError {
    cause: String,
}

impl std::error::Error for SchemaError {}

impl Display for SchemaError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "JSON doesn't match expected schema: {}", self.cause)
    }
}

/// All the possible errors that can occur while initializing [`SchemaInspector`]
#[derive(Debug, Clone, PartialEq, Eq)]
pub enum InitError {
    SchemaNotFound(String),
    ResolutionFailure(String), // contains the reference that was not resolved
    InvalidSchema(Option<String>), // contains the missing field in the schema
    CompileFailure,
    InferenceFailure,
}

impl std::error::Error for InitError {}

impl Display for InitError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let base = "failed to initialize Inspector:";
        match self {
            Self::SchemaNotFound(schema) => {
                write!(f, "{base} {schema} was not found in the provided schemas")
            }
            Self::ResolutionFailure(schema) => write!(
                f,
                "{base} failed to resolve references: referenced schema {schema} was not found"
            ),
            Self::InferenceFailure => write!(f, "{base} name of the type couldn't be inferred"),
            Self::CompileFailure => write!(f, "{base} failed to compile schema"),
            Self::InvalidSchema(schema) => write!(
                f,
                "{base} invalid schema provided: {} is missing or is not an object",
                schema.as_deref().unwrap_or("field")
            ),
        }
    }
}
